/*
   DX2VRML: 

   An outboard module for IBM Data Explorer V2.0 which converts an input
   object into the VRML file format.

   (c)1995 Christopher Kline <ckline@acm.org>
           Cornell Theory Center
	   Cornell University, Department of Computer Science

    [Release 1.2 <4-January-1996>]
    [Release 1.1 <1-August-1995>]
    [Release 1.0 <17-July-1995>]
*/

#include <dx/dx.h>
#include <stdio.h>
#include <math.h>
#include <string.h>
    
Error m_Dx2vrml(Object *in);
Error ExtractDXstructure(Object o, int IgnoreLighting, int IgnoreNormals, 
			 int IgnoreMaterials, int IgnoreRenderingProperties, 
			 int IgnoreOpacities);
Error tab();
void cleanup();

Error VRMLHeader(FILE *f);
Error VRMLFooter(FILE *f);
Error VRMLSeparator_Begin(FILE *f);
Error VRMLSeparator_End(FILE *f);
Error VRMLGroup_Begin(FILE *f);
Error VRMLGroup_End(FILE *f);
Error VRMLWWWAnchor_Begin(FILE *f, Object *in);
Error VRMLWWWAnchor_End(FILE *f);
Error VRMLCamera(FILE *f, Camera cam);
Error VRMLTransform(FILE *f, Matrix NetTransMatrix);
Error VRMLNormal(FILE *f, Array normarray, char *binding, int *nnormals);
Error VRMLCoordinate3(FILE *f, Array posarray);
Error VRMLIndexedFaceSet(FILE *f, Array connarray, char *conntype, 
			 char *normalsdep, char *colorsdep, int nnormals,
			 int ncolors, int IgnoreNormals, int IgnoreMaterials);
Error VRMLMaterial(FILE *f, Array colarray, Array opacarray, int *numcolors,
		   char *colorbinding, char *connectionstype, float ambient, 
		   float diffuse, float specular,int shininess,
		   int IgnoreOpacities, int IgnoreRenderingProperties );
Error VRMLShapeHints(FILE *f);
Error VRMLDirectionalLight(FILE *f, Light l);

Error WebspaceHints(FILE *f, Object *in);
Error WSTitleHints(FILE *f, Object *in);
Error WSCameraHints(FILE *f, Object *in);


/* 
   Macro to convert DX shininess values in the range 0-30 to VRML shininess
   values in the range 0-1, truncation if DX values fall outside either
   end of the range 
   */

#define MAXS 50
#define DXtoVRMLShininess(dx) (((double)dx)/MAXS)

/* An alternate definition. I'm not sure which one is better. */
/*
#define DXtoVRMLShininess(dx) (dx<1 ? 0.001:(dx>MAXS?1:log10(dx)/log10(MAXS))) 
*/

/* CREASEANGLE: If the interior angle (in radians) between two faces is larger 
   than CREASEANGLE, the browser should render a distinct edge where they meet;
   otherwise the browser should smooth-shade the edge */
#define CREASEANGLE "0.5" /* must be a string */

#define DX2VRML_VERSION "version 1.2"
#define SEPARATORS_AROUND_DX_FIELDS FALSE

#define FALSE 0
#define TRUE 1

#define BAD 0x2BAD
#define GOOD -BAD
#define INC 9999999 
#define DEC -9999999
#define RESET 0

#define TAB_INPUT 0
#define TAB_TITLE 1
#define TAB_URL 2
#define TAB_URLDESCRIPTION 3
#define TAB_URLMAP 4
#define TAB_FILENAME 5
#define TAB_IGNORE_LIGHTING 6
#define TAB_IGNORE_NORMALS 7
#define TAB_IGNORE_MATERIALS 8
#define TAB_IGNORE_OPACITIES 9
#define TAB_IGNORE_RENDERINGPROPERTIES 10
#define TAB_CAMSTART 11
#define MAX_NUM_TABS 22

FILE *file;

char *tabstr = NULL;
long number_of_polygons = 0;
int gotWWWAnchor = 0;

ErrorCode err;

/*---------------------------------------------------------------------*/

Error m_dx2vrml(Object *in) {
  char message[30], *greeting;
  Class theClass;
  char *f;
  char filename[150];
  int IgnoreLighting, IgnoreNormals, IgnoreMaterials, 
    IgnoreRenderingProperties, IgnoreOpacities;
  
  /* reset the error level so we can use this in cleanup() to see if
     an error occurred and we're jumping ship, or if we're just cleaning
     up after the module is done
     */

  DXMessage("");
  DXMessage("Entering DXVRML\n");
  DXMessage("");

  tabstr = (char *)NULL;
  tab(RESET); /* if tabstr isn't reset at the start of the code, it will 
		 corrupt the output file by printing a "(null)" before 
		 the first separator when run on SGI systems, because the
		 SGI version of fprintf() prints "(null)" if you try
		 printing a null string with %s */

  /* Check to see if we have any input, and if so, that it is the right kind */
  
  if (in[TAB_INPUT] == NULL) { 
    DXMessage("NO INPUT STRUCTURE!");
  }
  else {
    
    theClass = DXGetObjectClass(in[TAB_INPUT]);
    if ((theClass != CLASS_FIELD) 
	&& (theClass != CLASS_GROUP) 
	&& (theClass != CLASS_XFORM)
	&& (theClass != CLASS_LIGHT))
      DXErrorGoto(ERROR_BAD_CLASS, 
		  "Input must be a group, field, light, or Xform");
    

    /* Determine the output file name from input */
    if (in[TAB_FILENAME] != NULL) {
      if (DXExtractString(in[TAB_FILENAME],&f) == NULL)
	DXErrorReturn(ERROR_UNEXPECTED, "Output file name was not specified!")
      else {
	strcpy(filename, f);
	strcat(filename, ".wrl");
      } 
    } else DXErrorGoto(ERROR_BAD_PARAMETER, "Output File must be specified");

    /* Try to open the output file */
    DXMessage("Trying to open output file \"%s\"...", filename);
    if ((file = fopen(filename, "w")) == NULL) 
      DXErrorGoto(ERROR_UNEXPECTED, "Attempt to open output file failed");


    /* Try to create the VRML header */
    err = VRMLHeader(file);
    if (err == BAD) 
      DXErrorGoto(ERROR_UNEXPECTED, "Attempt to write VRML header failed");
    if (err != GOOD) goto error;

    DXMessage("Header written ok...");
    
    /* Insert Webspace-specific viewer hints */
    err = WebspaceHints(file, in);
    if (err == BAD) 
      DXErrorGoto(ERROR_UNEXPECTED, 
		  "Attempt to add hints for VRML viewer \"Webspace\" failed");
    if (err != GOOD) goto error;

    DXMessage("Webspace hints added...");

    /* Try to begin the VRML WWWAnchor node */
    gotWWWAnchor = 0;
    err = VRMLWWWAnchor_Begin(file, in);
    if (err == BAD) 
      DXErrorGoto(ERROR_UNEXPECTED, "Attempt to begin VRML WWWAnchor failed");
    if (err != GOOD) goto error;

    DXMessage("WWWAnchor begun ok...");

    /* Try to add shape hints so the viewer can optimize rendering */
    err = VRMLShapeHints(file);
    if (err == BAD) 
      DXErrorGoto(ERROR_UNEXPECTED,"Attempt to write VRML shape hints failed");
    if (err != GOOD) goto error;
    
    DXMessage("Shape hints added...");

    DXMessage("");

  /* Try to extract info on whether or not the user to ignore certain things */

    if (in[TAB_IGNORE_LIGHTING] != NULL) {
      if (DXExtractInteger(in[TAB_IGNORE_LIGHTING], &IgnoreLighting) == NULL) 
	DXErrorReturn(ERROR_UNEXPECTED, 
		      "Couldn't extract Ignore Lighting value")
      else if ((IgnoreLighting != 0) && (IgnoreLighting != 1))
	DXErrorReturn(ERROR_BAD_PARAMETER, 
		      "Ignore Lighting input value must be 0 or 1");
    }
    else {
      DXMessage("Ignoring Lighting as default.");
      IgnoreLighting = 1; /* default value */
    }
    
    if (in[TAB_IGNORE_NORMALS] != NULL) {
      if (DXExtractInteger(in[TAB_IGNORE_NORMALS], &IgnoreNormals) == NULL) 
	DXErrorReturn(ERROR_UNEXPECTED, 
		      "Couldn't extract Ignore Normals value")
      else if ((IgnoreNormals != 0) && (IgnoreNormals != 1))
	DXErrorReturn(ERROR_BAD_PARAMETER, 
		      "Ignore Normals input value must be 0 or 1");
    }
    else {
      DXMessage("NOT Ignoring Normals as default.");
      IgnoreNormals = 0; /* default value */
    }
 
    if (in[TAB_IGNORE_MATERIALS] != NULL) {
      if (DXExtractInteger(in[TAB_IGNORE_MATERIALS],&IgnoreMaterials) == NULL) 
	DXErrorReturn(ERROR_UNEXPECTED, 
		      "Couldn't extract Ignore Colors value")
      else if ((IgnoreMaterials != 0) && (IgnoreMaterials != 1))
	DXErrorReturn(ERROR_BAD_PARAMETER, 
		      "Ignore Colors input value must be 0 or 1");
    }
    else {
      DXMessage("NOT Ignoring Colors as default.");
      IgnoreMaterials = 0; /* default value */
    }
    
    if (in[TAB_IGNORE_OPACITIES] != NULL) {
      if (DXExtractInteger(in[TAB_IGNORE_OPACITIES],&IgnoreOpacities) == NULL) 
	DXErrorReturn(ERROR_UNEXPECTED, 
		      "Couldn't extract Ignore Opacities value")
      else if ((IgnoreOpacities != 0) && (IgnoreOpacities != 1))
	DXErrorReturn(ERROR_BAD_PARAMETER, 
		      "Ignore Opacities input value must be 0 or 1");
    }
    else {
      DXMessage("NOT Ignoring Opacities as default.");
      IgnoreOpacities = 0; /* default value */
    }
    
    if (in[TAB_IGNORE_RENDERINGPROPERTIES] != NULL) {
      if (DXExtractInteger(in[TAB_IGNORE_RENDERINGPROPERTIES], 
			   &IgnoreRenderingProperties) == NULL) 
	DXErrorReturn(ERROR_UNEXPECTED, 
		      "Couldn't extract Ignore Rendering Properties value")
      else if ((IgnoreRenderingProperties != 0) && 
	       (IgnoreRenderingProperties != 1))
	DXErrorReturn(ERROR_BAD_PARAMETER, 
		      "Ignore Rendering Properties input must be 0 or 1");
    }
    else {
      DXMessage("Ignoring Rendering Properties as default.");
      IgnoreRenderingProperties = 0; /* default value */
    }
    

    /* Reset any errors so that our function return values are interpreted 
       correctly */
    DXResetError();
    
    /* Now proceed in extracting the guts of the DX object */
    number_of_polygons = 0;
    err = ExtractDXstructure(in[TAB_INPUT], IgnoreLighting, IgnoreNormals, 
			     IgnoreMaterials, IgnoreRenderingProperties, 
			     IgnoreOpacities);
    if (err == BAD) 
      DXErrorGoto(ERROR_UNEXPECTED, "Attempt to write VRML scene failed while traversing the input object.");
    if (err != GOOD) goto error;
    
    /* Try to end the VRML WWWAnchor node */
    if (gotWWWAnchor) {
      err = VRMLWWWAnchor_End(file);
      if (err == BAD) 
	DXErrorGoto(ERROR_UNEXPECTED, "Attempt to end VRML WWWAnchor failed");
      if (err != GOOD) goto error;
    }
    
      DXMessage("WWWAnchor ended ok...");

    err = VRMLFooter(file);
    if (err == BAD) 
      DXErrorGoto(ERROR_UNEXPECTED, "Attempt to write VRML footer failed");
    if (err != GOOD) goto error;

  }
  
 error:
  cleanup();
  
  
  DXMessage("");
  DXMessage("Approximate number of polygons in output: %ld",
	    number_of_polygons);
  DXMessage("");
  DXMessage("Leaving DX2VRML");
  DXMessage("");

  return(OK);
}


/*---------------------------------------------------------------------*/

Error ExtractDXstructure(Object o, int IgnoreLighting, int IgnoreNormals, 
			 int IgnoreMaterials, int IgnoreRenderingProperties, 
			 int IgnoreOpacities) {
     
  Array a,b,c,d,e;	     /* Used as refrences to data arrays.*/
  Object tempobj;	       	     /* temporary object to recurse on */
  char *normalsdep, *colorsdep, *opacitiesdep, *connectiontype;
  int temp, npos,ncons;   
  Matrix matrix;		     /* Matrix used during Xform extraction */
  float T[16];	                /* Temp storage used during Xform extraction */
  Class theClass;
  float ambient, diffuse, specular;  /* rendering modifiers */  
  int shininess;                     /* rendering modifier */
  int numnormals, numcolors;
  
  /* Determine what the object is */
  theClass = DXGetObjectClass(o);
  
  switch (theClass){
    
  case CLASS_XFORM:
    
    /* Extract the Xform information and concatenate
     * into one net Transformation Matrix which is stored in T
     * to be used to transform the object vertices later.
     */
    
    DXMessage("");
    DXMessage("The data is an XFORM\n");
    
    
    DXMessage("Got an Xform");
    if (DXGetXformInfo((Xform)o,&tempobj,&matrix) == NULL) return ERROR;

    /* Try to create the VRML transformation */

    if (tempobj != NULL) {
      if ((err = VRMLSeparator_Begin(file)) != GOOD) return(err);
    }

    if ((err = VRMLTransform(file, matrix)) != GOOD) return(err);

    if (tempobj != NULL){
      DXMessage("Recursing to extract the stuff being transformed by XForm");

      if ((err = ExtractDXstructure(tempobj, IgnoreLighting, 
				    IgnoreNormals, IgnoreMaterials, 
				    IgnoreRenderingProperties, 
				    IgnoreOpacities)) != GOOD) return(err);
      if ((err = VRMLSeparator_End(file)) != GOOD) return(err);
    }

    break;
    
  case CLASS_LIGHT:

    if (IgnoreLighting == FALSE) {

      DXMessage("");
      DXMessage("The data is a LIGHT\n");
      
      if ((err = VRMLDirectionalLight(file, (Light)o)) != GOOD) return(err);
    
    }
    
    break;

  case CLASS_FIELD:
    
    /*
     *Get the Postions, Connections, normals, attributes, and colors
     */

    DXMessage("");
    DXMessage("The data is a FIELD\n");

    /* ----- get the positions -----*/

    DXMessage("Attempting to grab positions...");

    if ((a=(Array) DXGetComponentValue((Field)o,"positions")) == NULL) {
      /* This is bad. Can't see an object with no positions! */
      DXWarning("This field has no positions. Ignoring this field.");
      return(GOOD);
    }

    if (SEPARATORS_AROUND_DX_FIELDS) {       
      if ((err = VRMLSeparator_Begin(file)) != GOOD) return(err);
    }

    if ((err = VRMLCoordinate3(file, a)) != GOOD) return(err);
    
    /* ----- Ok, now let's grab the connections ----- */
    
    DXMessage("");
    DXMessage("Attempting to grab connections...");
    
    
    if ((b = (Array) DXGetComponentValue((Field)o,"connections")) == NULL) {
      DXWarning("This field has no connections. Ignoring connections.");

      /* no connections, so no faces, so no faces to shade -> NO NORMALS! :) */
      
      IgnoreNormals = TRUE;

      /* grab some memory to put text into, since DX didn't do it for us.
	 It took me a REALLY long time to find that bug ;) */
      if ((connectiontype = (char *)DXAllocateZero(10*sizeof(char))) == NULL)
	return(BAD);
      strcpy(connectiontype, "POINTSET"); 

      DXMessage("Setting connectiontype to %s", connectiontype);

    } else {
      if (DXExtractString(DXGetComponentAttribute((Field)o, "connections", "element type" ), &connectiontype) == NULL) {
	DXErrorReturn(ERROR_UNEXPECTED, "Couldn't get connections type");
      }
    }
    
    DXMessage("Got the connections type, and it is [%s]", connectiontype);
    
    if ((strcmp(connectiontype,"tetrahedra") == 0) || 
	(strcmp(connectiontype,"cubes") == 0)) {
      /* Ignore normals because DX's default renderer doesn't use normals
	 when volume rendering, and if normals ARE present (because someone
	 added them to support someone else's renderer), we won't know
	 how to handle them correctly since we fake volume rendering by
	 converting cube/tetra connections to 6/4 separate faces */
      IgnoreNormals = TRUE;
      DXMessage("Tetra or Cube connections. We're going to ignore normals!");
    }


    /* ----- get the normals ----- */

    if (IgnoreNormals == FALSE) {
      /* if connections aren't lines (lines don't have normals), get normals */
      if (strcmp(connectiontype,"lines") != 0) { 
	DXMessage("");
	DXMessage("Attempting to grab normals...");
	
	if ((d = (Array)DXGetComponentValue((Field)o,"normals")) == NULL) {
	  IgnoreNormals = TRUE; /* set IgnoreNormals so that a normalIndex 
				   field is not added to the IndexedFaceSet
				   node */
	  DXMessage("This field has no normals. Ignoring normals.");
	}
	else {
	  
	  if (DXExtractString(DXGetComponentAttribute((Field)o,
						      "normals","dep"), 
			      &normalsdep) == NULL) {
	    DXErrorReturn(ERROR_UNEXPECTED,"Couldn't get normals dependency");
	    return(ERROR);
	  }
	  
	  DXMessage("Got the normals dependency, and it is [%s]", normalsdep);
	  
	  /* Try to create the VRML normals list */
	  if ((err = VRMLNormal(file, d, normalsdep, &numnormals)) != GOOD) 
	    return(err);
	}
      }
  } 
    
    /* ----- get the colors ----- */
  

    if (IgnoreMaterials == FALSE) {
      
      DXMessage("");
      DXMessage("Attempting to grab colors...");
      
      /* check for front colors */
      if ((c = (Array)DXGetComponentValue((Field)o,"front colors")) == NULL) {
	DXMessage("This field has no \"front colors\".Looking for \"colors\"");
	
	/* check for colors */
	if ((c = (Array)DXGetComponentValue((Field)o,"colors")) == NULL) {
	  DXMessage("This field has no \"colors\". Ignoring colors");
	  IgnoreMaterials = TRUE; /* ignore the materials */
	}
	else {
	  DXMessage("Got \"colors\"");
	  DXMessage("Using \"colors\"");
	}    
      }
      else {
	DXMessage("Got \"front colors\"");
	DXMessage("Using \"front colors\"");
      }
      

      if (DXExtractString(DXGetComponentAttribute((Field)o, "colors","dep"),
			  &colorsdep) == NULL) {
	DXMessage("Couldn't get colors dependency, Ignoring Materials");
	  IgnoreMaterials = TRUE; /* ignore the materials */	
      }
      
      DXMessage("Got the colors dependency, and it is [%s]", colorsdep);
      
      if ((strcmp(connectiontype,"POINTSET") == 0) && 
	  (strcmp(colorsdep,"connections") == 0)) {
	DXErrorReturn(ERROR_INVALID_DATA, "This field contains colors that are dependent upon connections, but there are not any connections present!");
      }

      if (IgnoreOpacities == FALSE) {
	
	/* try to get opacities */
	DXMessage("Attempting to grab opacities...");
	
	if ((e=(Array) DXGetComponentValue((Field)o,"opacities")) == NULL) { 
	  DXMessage("This field has no opacities. Ignoring opacities.");
	  IgnoreOpacities == TRUE;
	}
	else {
	  if (DXExtractString(DXGetComponentAttribute((Field)o, "opacities",
						      "dep"),&opacitiesdep) 
	      == NULL) {
	    DXErrorReturn(ERROR_UNEXPECTED,"Couldn't get opacities dependency");
	  }
	  else {
	    DXMessage("Got the opacities dependency, and it's [%s]",
		      opacitiesdep);
	    if (strcmp(opacitiesdep, colorsdep) != 0) {
	      DXWarning("Colors and opacities have different dependencies!");
	      DXWarning("Ignoring opacities.");
	      IgnoreOpacities = TRUE;
	    }	    
	  }
	  
	}
      } /* end -- if (IgnoreOpacities == FALSE) */
      
      /* are we IGNORING rendering properties? */
      if (IgnoreRenderingProperties == FALSE) {
	
	DXMessage("Attempting to extract material properties...");
	
	/* try to extract ambient, diffuse, specular, and shininess*/
	
	/* ambient */
	if (DXExtractFloat(DXGetAttribute(o, "front ambient"), &ambient) 
	    == NULL) {
	  DXMessage("Couldn't get \"front ambient\" attribute. Looking for \"ambient\"");
	  if (DXExtractFloat(DXGetAttribute(o, "ambient"), &ambient) == NULL) {
	    DXMessage("Couldn't get \"ambient\" attribute. Defaulting to 1");
	    ambient = 1;
	  }
	  else DXMessage("Got \"ambient\" attribute, value is %f", ambient);
	}
	else DXMessage("Got \"front ambient\" attribute, value is %f",ambient);
	
	/* diffuse */
	if (DXExtractFloat(DXGetAttribute(o, "front diffuse"), &diffuse) 
	    == NULL) {
	  DXMessage("Couldn't get \"front diffuse\" attribute. Looking for \"diffuse\"");
	  if (DXExtractFloat(DXGetAttribute(o, "diffuse"), &diffuse) == NULL) {
	    DXMessage("Couldn't get \"diffuse\" attribute. Defaulting to 0.7");
	    diffuse = 0.7;
	  }
	  else DXMessage("Got \"diffuse\" attribute, value is %f", diffuse);
	}
	else DXMessage("Got \"front diffuse\" attribute, value is %f",diffuse);
	
	/* specular */
	if (DXExtractFloat(DXGetAttribute(o, "front specular"), &specular)
	    ==NULL) {
	  DXMessage("Couldn't get \"front specular\" attribute. Looking for \"specular\"");
	  if (DXExtractFloat(DXGetAttribute(o, "specular"), &specular) 
	      == NULL) {
	    DXMessage("Couldn't get \"specular\" attribute.Defaulting to 0.5");
	    specular = 0.5;
	  }
	  else DXMessage("Got \"specular\" attribute, value is %f", specular);
	}
	else DXMessage("Got \"front specular\" attribute, value is %f", specular);
	
	/* shininess */
	if (DXExtractInteger(DXGetAttribute(o, "front shininess"), &shininess) 
	    == NULL) {
	  DXMessage("Couldn't get \"front shininess\" attribute. Looking for \"shininess\"");
	  if (DXExtractInteger(DXGetAttribute(o, "shininess"), &shininess)
	      ==NULL) {
	    DXMessage("Couldn't get \"shininess\" attribute.Defaulting to 10");
	    shininess = 10;
	  }
	  else 
	    DXMessage("Got \"shininess\" attribute, value is %d", shininess);
	}
	else DXMessage("Got \"front shininess\" attribute, value is %d",
		       shininess);
	
      } /* end - if (IgnoreRenderingProperties == FALSE) { */      
      

      /* check again in case we had a problem getting colors and turned materials off inside 
	 the first "if (IgnoreMaterials == FALSE)" check */
      if (IgnoreMaterials == FALSE) { 
	
	/* output the colors and attributes */
	if ((err = VRMLMaterial(file, c, e, &numcolors, colorsdep, 
				connectiontype, ambient, diffuse,specular,
				shininess, IgnoreOpacities, 
				IgnoreRenderingProperties)) != GOOD) return(err);

      } /* end - if (IgnoreMaterials == FALSE) { */


    }  /* end - if (IgnoreMaterials == FALSE) { */ 
    

    /* NOW make the face set */
    if ((err = VRMLIndexedFaceSet(file, b, connectiontype, normalsdep,
				  colorsdep, numnormals, numcolors,
				  IgnoreNormals, IgnoreMaterials)) 
	!= GOOD) return(err);
    
    if (SEPARATORS_AROUND_DX_FIELDS) {
      if ((err = VRMLSeparator_End(file)) != GOOD) return(err);
    }


    break;
    


  case CLASS_GROUP:

    /* Recursively traverse Groups */

    DXMessage("");
    DXMessage("The data is a GROUP\n");

    DXMessage("Recursing on group to get rest of structure");

    if ((err = VRMLSeparator_Begin(file)) != GOOD) return(err);
    
    /* Take care of all lighting objects first, because if a light appears
       after an object in a VRML group node, it won't affect the object. 
       VRML lights need to appear BEFORE the objects they affect. This is
       stupid, but ain't that America? */
    for (temp=0;tempobj=DXGetEnumeratedMember((Group)o,temp,NULL);temp++) {
      if (DXGetObjectClass(tempobj) == CLASS_LIGHT) 
	if ((err = ExtractDXstructure(tempobj, IgnoreLighting, 
				      IgnoreNormals, IgnoreMaterials, 
				      IgnoreRenderingProperties, 
				      IgnoreOpacities)) != GOOD) return(err);
    }
    
    /* Now take care of all the things that aren't lights. I KNOW this is a
       stupid way to handle the VRML light problem, but I'm tired and don't 
       want to bother building a linked list of objects in the correct order,
       or some other intelligent data structure. */
    for (temp=0;tempobj=DXGetEnumeratedMember((Group)o,temp,NULL);temp++) {
      if (DXGetObjectClass(tempobj) != CLASS_LIGHT)
	if ((err = ExtractDXstructure(tempobj, IgnoreLighting, 
				      IgnoreNormals, IgnoreMaterials, 
				      IgnoreRenderingProperties, 
				      IgnoreOpacities)) != GOOD) return(err);
    }

    if ((err = VRMLSeparator_End(file)) != GOOD) return(err);
    
    break;
    
  default:
    
    DXMessage("");
    DXWarning("Object is not a group, field, light, or Xform. Ignoring it.");
    return(GOOD);
    break;
    
  } /*End Switch*/
    

  DXMessage("");
  DXMessage("Exiting Recursion Normally");

  /* Shouldn't be any errors at this point */
  return(GOOD);

}

/*---------------------------------------------------------------------*/

void cleanup() {

  char *errorstring;

  if ((err = DXGetError()) != ERROR_NONE){
    errorstring = (char *)strdup(DXGetErrorMessage());
    DXWarning("   An Error Occurred, better clean up before we jump ship.");
    DXWarning("   The error was [%d] \"%s\"", err, errorstring);
    DXWarning("   The output file may be corrupt!");
    DXMessage("");
  }
  

  DXMessage("Cleaning up...");
  DXMessage("");
  DXMessage("Closing output file");

  fclose(file);

}

/*---------------------------------------------------------------------*/

Error tab(int adjust) {

 static int numtabs = 0;
 int i;

 
 if (adjust == INC) numtabs++;
 else 
   if (adjust == DEC) numtabs--;
   else 
     numtabs = adjust;
 
 if (tabstr != NULL) DXFree(tabstr);
 
 if ((tabstr = (char *)DXAllocateZero((numtabs+1)*sizeof(char))) == NULL)
   return(BAD);

 for (i=0; i<numtabs; i++)
   tabstr = strcat(tabstr, "\t");
 
 return(GOOD);

}

/*---------------------------------------------------------------------*/

Error VRMLHeader(FILE *f) {

  fprintf(f, "#Inventor V2.0 ascii\n");
  if ((err = VRMLSeparator_Begin(file)) != GOOD) return(err);

  return(GOOD);
}


/*---------------------------------------------------------------------*/

Error VRMLFooter(FILE *f) {

  
  if (fprintf(f, "%sInfo {\n", tabstr) < 0) return(BAD);
  tab(INC);
  fprintf(f, "%sstring \"This file was created using Dx2vrml, a module for the\"\n",tabstr);
  fprintf(f, "%sstring \"IBM Data Explorer scientific visualization software\"\n",tabstr);
  fprintf(f, "%sstring \"package.\"\n",tabstr);
  fprintf(f, "%sstring \"(c)1995 Christopher Kline <ckline@acm.org>\"\n",tabstr);
  fprintf(f, "%sstring \"        Cornell University, Dept. of Computer Science\"\n\n\n",tabstr);

  fprintf(f, "%sstring \"Version of Dx2vrml used: %s\"\n",tabstr, DX2VRML_VERSION);
  fprintf(f, "%sstring \"Number of explicitly defined polygons: %ld\"\n",tabstr, number_of_polygons);
  
  tab(DEC);
  fprintf(f, "%s}\n",tabstr);

  if ((err = VRMLSeparator_End(file)) != GOOD) return(err);

  return(GOOD);
}

/*---------------------------------------------------------------------*/

Error VRMLGroup_Begin(FILE *f) {

  if (fprintf(f, "%sGroup {\n", tabstr) < 0)
    return(BAD);
  
  tab(INC);
     
  return(GOOD);
}

/*---------------------------------------------------------------------*/

Error VRMLGroup_End(FILE *f) {

  return(VRMLSeparator_End(f));

}
/*---------------------------------------------------------------------*/

Error VRMLSeparator_Begin(FILE *f) {

  if (fprintf(f, "%sSeparator {\n", tabstr) < 0)
    return(BAD);
  
  tab(INC);
     
  return(GOOD);
}

/*---------------------------------------------------------------------*/

Error VRMLSeparator_End(FILE *f) {

  tab(DEC);

  if (fprintf(f, "%s}\n", tabstr) < 0)
    return(BAD);
  
  return(GOOD);
}


/*---------------------------------------------------------------------*/

Error VRMLWWWAnchor_Begin(FILE *f, Object *in) {

  char *url, *desc;
  char domap[10];
  int mapval;

  /* try to extract a URL */
  if (in[TAB_URL] != NULL) {
    if (DXExtractString(in[TAB_URL], &url) == NULL) 
      DXErrorReturn(ERROR_UNEXPECTED, "Couldn't extract URL of scene")
    else {
      /* got a URL */
      if (strstr(url, "://") == NULL) {
	DXErrorReturn(ERROR_UNEXPECTED, "URL specified is not a valid URL");
      }
      gotWWWAnchor++;
      if (fprintf(f, "%sWWWAnchor {\n", tabstr) < 0) return(BAD);
      tab(INC);
      fprintf(f, "%sname\t\"%s\"\n", tabstr, url);        
    }
  }
  else {
    DXMessage("No scene URL was specified. Ignoring URL.");
    return(GOOD);
  }

  /* try to extract a URL description */
  if (in[TAB_URLDESCRIPTION] != NULL) {
    if (DXExtractString(in[TAB_URLDESCRIPTION], &desc) == NULL) 
      DXErrorReturn(ERROR_UNEXPECTED, "Couldn't extract URL description")
    else {
      /* got a description */
      fprintf(f, "%sdescription\t\"%s\"\n", tabstr, desc);  
    }
  }
  else {
    DXMessage("No URL description was specified. Creating empty URL description.");
    fprintf(f, "%sdescription\t\"\"\n", tabstr); 
  }

  /* Try to extract info on whether or not the user wants URL mapping */

  if (in[TAB_URLMAP] != NULL) {
    if (DXExtractInteger(in[TAB_URLMAP], &mapval) == NULL) 
      DXErrorReturn(ERROR_UNEXPECTED, "Couldn't extract URL mapping value")
    else {
      /* got mapping info */
      if ((mapval != 0) && (mapval != 1))
	DXErrorReturn(ERROR_BAD_PARAMETER, 
		      "URL Mapping input value must be 0 or 1");
      fprintf(f, "%smap\t%s\n", tabstr, (mapval == 1) ? "POINT" : "NONE"); 
    }
  }
  else 
    fprintf(f, "%smap\t%s\n", tabstr, "NONE");  /* default value */
  
  return(GOOD);  
}

/*---------------------------------------------------------------------*/

Error VRMLWWWAnchor_End(FILE *f) {

  if (gotWWWAnchor > 0) {
    gotWWWAnchor--;
    return(VRMLSeparator_End(f));
  }
  else return(GOOD);
}

/*---------------------------------------------------------------------*/

Error VRMLCamera(FILE *f, Camera cam) {

  char camstr[30], heightstr[30];
  Point from, to;
  Vector up;
  float fov, aspect, width;
  Matrix m;
  double T[12];
  double theta;
  double axisx, axisy, axisz;
  double val;
  float temp;

  if (cam == NULL) {
   DXMessage("No camera input. Ignoring Cameras.");
   return(GOOD);
 }
  
  if (DXGetObjectClass((Object)cam) != CLASS_CAMERA) 
    DXErrorReturn(ERROR_BAD_CLASS, "Camera input is not of class Camera!");
  
  
  err = DXGetError();
  if (DXGetPerspective(cam, &fov, &aspect) == NULL) {
    if (err != DXGetError()) { 
      return(ERROR); /* must have been invalid camera since error was set */
    }
    else 
      /* since error wasn't set, DXGetPerspective() probably returned NULL 
	 because the camera was an orthographic camera and not a perspective 
	 camera */
      if (DXGetOrthographic(cam, &width, &aspect) == NULL) { 
	/* Ok, now we DEFINITELY have problems */
	return(ERROR);
      }
      else {
	strcpy(camstr, "OrthographicCamera");	
	strcpy(heightstr, "height");
	val = width * aspect;
	DXMessage("Got ORTHOGRAPHIC CAMERA");
	DXMessage("width = %g aspect = %g HEIGHT = %g", width,aspect, val);
      }
  }
  else {
    /* it was a perspective camera */
    strcpy(camstr, "PerspectiveCamera");
    strcpy(heightstr, "heightAngle");
    val = 2*atan(fov/2);
    DXMessage("Got PERSPECTIVE CAMERA"); 
    DXMessage("fov = %g aspect = %g ANGLE = %g", fov, aspect, val);
  }
  
  if (DXGetView(cam, &from, &to, &up) == NULL) 
    return(ERROR);
  else { 
    DXMessage("");
    DXMessage("  from = [%g, %g, %g]", from.x, from.y, from.z);
    DXMessage("  to = [%g, %g, %g]", to.x, to.y, to.z);
    DXMessage("  up = [%g, %g, %g]", up.x, up.y, up.z);
    DXMessage("");
  }

  m = DXGetCameraRotation(cam); /* apparently there's no way to error check
				   the result of this function call */
  
  T[0] = m.A[0][0];
  T[1] = m.A[0][1];
  T[2] = m.A[0][2];
  
  T[3] = m.A[1][0];
  T[4] = m.A[1][1];
  T[5] = m.A[1][2];

  T[6] = m.A[2][0];
  T[7] = m.A[2][1];
  T[8] = m.A[2][2];

  T[9] = m.b[0];
  T[10] = m.b[1];
  T[11] = m.b[2];

  DXMessage("Camera Transformation matrix is:");
  DXMessage("%g %g %g", T[0], T[1], T[2]);
  DXMessage("%g %g %g", T[3], T[4], T[5]);
  DXMessage("%g %g %g", T[6], T[7], T[8]);
  DXMessage("%g %g %g", T[9], T[10], T[11]);

  /* Calculate the axis and degree of rotation from the T[] matrix. 
     This algorithm is by Michael E. Pique, Research Institute of Scripps 
     Clinic,La Jolla, California. 
     Found in Graphics Gems (volume 1) page 465 */

  theta = acos( (T[0] + T[4] + T[8] - 1)/2 );
  
  if (sin(theta) == 0) {
    /* since sin(theta) == 0 we can't determing the axis of rotation */
    /* So we make a null rotation */
    DXMessage("Couldn't extract axis and rotation from camera matrix");
    axisx = axisy = axisz = 1;
    theta = 0;
  }
  else {
    axisx = (T[5] - T[7]) / (2 *sin(theta));
    axisy = (T[6] - T[2]) / (2 *sin(theta));
    axisz = (T[1] - T[3]) / (2 *sin(theta));
  }
  
  /* done calculating axis and degree of rotation */

  DXMessage("%srotation basis (%g, %g, %g)  %g radians\n", 
	    tabstr, axisx, axisy, axisz, theta);

  if (fprintf(f, "%s%s {\n", tabstr, camstr) < 0) return(BAD);
  tab(INC);

  /* reverse the angle of rotation */
  theta = -theta;

  /* normalize the components of the rotation basis */
  temp = sqrt(axisx*axisx + axisy*axisy +axisz*axisz);
  axisx = axisx/temp;
  axisy = axisy/temp;
  axisz = axisz/temp;

  /* compute the translation */
  T[9] = from.x;
  T[10] = from.y;
  T[11] = from.z;

  fprintf(f, "%sposition\t%g %g %g\n",tabstr, T[9], T[10], T[11]);
  fprintf(f, "%sorientation\t%g %g %g  %g\n",tabstr,axisx,axisy,axisz,theta);
  if (fprintf(f, "%s%s\t%g\n",tabstr, heightstr, val) < 0) return(BAD);

  tab(DEC);
  if (fprintf(f, "%s}\n",tabstr) < 0) return(BAD);
  
  return(GOOD);
}

/*---------------------------------------------------------------------*/

Error VRMLTransform(FILE *f, Matrix NetTransMatrix) {

  double T[12], val;
  char optype[200] = "unknown";
  char opstr[300] = "no op";

  T[0] = NetTransMatrix.A[0][0];
  T[1] = NetTransMatrix.A[0][1];
  T[2] = NetTransMatrix.A[0][2];
  
  T[3] = NetTransMatrix.A[1][0];
  T[4] = NetTransMatrix.A[1][1];
  T[5] = NetTransMatrix.A[1][2];

  T[6] = NetTransMatrix.A[2][0];
  T[7] = NetTransMatrix.A[2][1];
  T[8] = NetTransMatrix.A[2][2];

  T[9] = NetTransMatrix.b[0];
  T[10] = NetTransMatrix.b[1];
  T[11] = NetTransMatrix.b[2];

  DXMessage("%g %g %g", T[0], T[1], T[2]);
  DXMessage("%g %g %g", T[3], T[4], T[5]);
  DXMessage("%g %g %g", T[6], T[7], T[8]);
  DXMessage("%g %g %g", T[9], T[10], T[11]);

  if (!T[1] && !T[2] && !T[3] && !T[5] && !T[6] && 
      !T[7] && !T[9] && !T[10] && !T[11]) {
    /* (Most likely) a scaling. If it's not a scale, it's a rotation of zero
       degrees, or a translation of 0. They just happen to LOOK like scales*/
    if ((T[0] == 1) && (T[4] == 1) && (T[8] == 1)) {
      /* It's a scale of 1, or a rotation of 0 degrees, so just ignore it */
      DXWarning("XFORM was a homogenous scale of factor 1, a rotation of 0");
      DXWarning("degrees, or a homogenous translation of 0. Ignoring it."); 
      return(GOOD);
    }
    
    DXMessage("XFORM is a SCALE");
    sprintf(optype, "Scale");
    sprintf(opstr, "scaleFactor\t%g %g %g",T[0], T[4], T[8]);
  }
  else 
    if (!T[1] && !T[2] && !T[3] && !T[5] && !T[6] && !T[7] && 
	 (T[0] == 1) && (T[4] == 1) && (T[8] == 1))  {
      /* Must be a translation */
      DXMessage("XFORM is a TRANSLATION");
      sprintf(optype, "Translation");
      sprintf(opstr, "translation\t%g %g %g",T[9], T[10], T[11]);
    }
  else if (!T[9] && !T[10] && !T[11]) {
    /* At this point, it must be a rotation */
    DXMessage("XFORM is a ROTATION");

    sprintf(optype, "Rotation");
    
    if ((T[4]==T[8]) && !T[1] && !T[2] && !T[3] && !T[6] && (T[0]==1)) /* Rx */
      sprintf(opstr, "rotation\t1 0 0  %g", acos(T[8]));
    else if ((T[0]==T[8]) && !T[1] && !T[3] &&!T[5]&&!T[7]&&(T[4]==1)) /* Ry */
      sprintf(opstr, "rotation\t0 1 0  %g", acos(T[8]));
    else if ((T[0]==T[4]) && !T[2] && !T[5] &&!T[6]&&!T[7]&&(T[8]==1)) /* Rz */
      sprintf(opstr, "rotation\t0 0 1  %g", acos(T[4]));
  }
  else {
    DXMessage("XFORM is an ARBITRARY MATRIX TRANSFORMATION");
    sprintf(optype, "MatrixTransform");
    sprintf(opstr, "matrix\t%g %g %g 0  %g %g %g 0  %g %g %g 0  %g %g %g 1",
	    T[0], T[1], T[2], T[3], T[4], T[5], T[6], T[7], T[8],
	    T[9], T[10], T[11]);      
  }
  
  if (fprintf(f, "%s%s {\n", tabstr, optype) < 0) return(BAD);
  tab(INC);
  fprintf(f, "%s%s\n", tabstr, opstr);
  tab(DEC);
  if (fprintf(f, "%s}\n",tabstr) < 0) return(BAD);
  return(GOOD);
}

/*---------------------------------------------------------------------*/

Error VRMLNormal(FILE *f, Array normarray, char *binding, int *nnormals) {
  
  ArrayHandle h;
  int nitems, rank, shape, i;
  Type type;
  Category category;
  Vector *data;
  Vector scratch;
  char bindstr[25];

  if ((h =  DXCreateArrayHandle(normarray)) == NULL) return(ERROR);

  if (DXGetArrayInfo(normarray,&nitems,&type,&category,&rank,&shape) == NULL) 
    DXErrorReturn(ERROR_UNEXPECTED,
		  "Could not get information about the array of normals");

  DXMessage("Got %d normals", nitems);
  
  /* set the number of normals in case it's needed */
  if (nnormals != NULL) *nnormals = nitems;

  if ((rank != 1) || (shape != 3) || 
      (type != TYPE_FLOAT) || (category != CATEGORY_REAL)) 
    DXErrorReturn(ERROR_UNEXPECTED,
		  "Normals are not an array of floating point 3-vectors");


  /* Now output the normal bindings (i.e., are normals bound to vertices
     or faces?) 
     */

  if (strcmp(binding,"connections") == 0) {
    sprintf(bindstr, "PER_FACE");
    DXMessage("Setting normal binding to PER_FACE");
  }
  else 
    if (strcmp(binding,"positions") == 0) {
      sprintf(bindstr, "PER_VERTEX_INDEXED");
      DXMessage("Setting normal binding to PER_VERTEX_INDEXED");
    }
  else {
    DXWarning("Unusual normal binding. I'm guessing by using PER_FACE");
    sprintf(bindstr, "PER_FACE");
    DXMessage("Setting normal binding to PER_FACE");
  }

  if (fprintf(f, "%sNormalBinding {\n", tabstr) < 0) return(BAD);
  tab(INC);
  fprintf(f, "%svalue\t%s\n", tabstr, bindstr);
  tab(DEC);
  if (fprintf(f, "%s}\n",tabstr) < 0) return(BAD);


  /* Output the normals */

  if (fprintf(f, "%sNormal {\n", tabstr) < 0) return(BAD);
  tab(INC);
  fprintf(f, "%svector [", tabstr);

  for (i = 0; i < nitems; i++) {
    data = DXGetArrayEntry(h, i, &scratch);
    fprintf(f, "%s\t%g %g %g,\n", tabstr, data->x, data->y, data->z);
}
  
  fprintf(f, "%s]\n", tabstr);
  tab(DEC);
  if (fprintf(f, "%s}\n",tabstr) < 0) return(BAD);



  return(GOOD);
}

/*---------------------------------------------------------------------*/
Error VRMLCoordinate3(FILE *f, Array posarray) {
  
  ArrayHandle h;
  int nitems, rank, shape, i;
  Type type;
  Category category;
  Vector *data;
  Vector scratch;

  if ((h =  DXCreateArrayHandle(posarray)) == NULL) return(ERROR);

  if (DXGetArrayInfo(posarray,&nitems,&type,&category,&rank,&shape) == NULL)
    DXErrorReturn(ERROR_UNEXPECTED,
		  "Could not get information about the array of positions");

  DXMessage("Got %d positions", nitems);

  if ((rank != 1) || ((shape != 2) && (shape != 3)) || 
      (type != TYPE_FLOAT) || (category != CATEGORY_REAL))
    DXErrorReturn(ERROR_UNEXPECTED,
		  "Positions are not an array of floating point 2- or 3-vectors");

  /* Output the positions*/

  if (fprintf(f, "%sCoordinate3 {\n", tabstr) < 0) return(BAD);
  tab(INC);
  fprintf(f, "%spoint [", tabstr);

  for (i = 0; i < nitems; i++) {
    data = DXGetArrayEntry(h, i, &scratch);
    if (shape == 2)
      fprintf(f, "%s\t%g %g 0,\n", tabstr, data->x, data->y);
    else
      fprintf(f, "%s\t%g %g %g,\n", tabstr, data->x, data->y, data->z);
  }
  
  fprintf(f, "%s]\n", tabstr);
  tab(DEC);
  if (fprintf(f, "%s}\n",tabstr) < 0) return(BAD);

  return(GOOD);
}

/*---------------------------------------------------------------------*/

Error VRMLIndexedFaceSet(FILE *f, Array connarray, char *conntype, 
			 char *normalsdep, char *colorsdep, int nnormals,
			 int ncolors, int IgnoreNormals, int IgnoreMaterials) {
  
  ArrayHandle h;
  int nitems, rank, shape, i;
  Type type;
  Category category;
  Cube scratch;   /* We only handle elements with 2, 3, 4, or 8 connections --
		     nothing larger! */
  Cube *data; 

  if ((h =  DXCreateArrayHandle(connarray)) == NULL) return(ERROR);

  if (strcmp(conntype,"POINTSET") == 0) { /*a point set */
    /* handle the special case where there are positions but no connections */
    DXMessage("Handling POINTSET");
    if (fprintf(f, "%sPointSet {\n", tabstr) < 0) return(BAD);
    fprintf(f, "%s\tstartIndex\t0\n%s\tnumPoints\t-1\n", tabstr, tabstr);
    goto skipconnectionstuff;
  }

  if (DXGetArrayInfo(connarray,&nitems,&type,&category,&rank,&shape)== NULL) 
    DXErrorReturn(ERROR_UNEXPECTED,
		  "Could not get information about the array of connections");
  
  DXMessage("Got %d connections", nitems);

  if ((rank != 1) || (type != TYPE_INT) || (category != CATEGORY_REAL) ||
      ((shape != 2) && (shape != 3) && (shape != 4) && (shape != 8)))
    DXErrorReturn(ERROR_UNEXPECTED,
		  "Connections are not an array of integer [2,3,4,8]-vectors");

  /* Output the connections */

  if (strcmp(conntype,"lines") == 0) { 
    if (fprintf(f, "%sIndexedLineSet {\n", tabstr) < 0) return(BAD);
  }
  else
    if (fprintf(f, "%sIndexedFaceSet {\n", tabstr) < 0) return(BAD);

  tab(INC);
  fprintf(f, "%scoordIndex [", tabstr);
  
  for (i = 0; i < nitems; i++) {
    data = DXGetArrayEntry(h, i, &scratch);
    
    /* Note: DX stores connections in a strange order, while VRML expects
       something simple, like counter-clockwise. So carefully note the order
       we output the vertex numbers in the coordIndex field!
       */
    
    if (strcmp(conntype,"lines") == 0) {
      fprintf(f, "%s\t%d, %d, -1,\n", tabstr, data->p, data->q);
      
      /* This is really browser-dependent, but figure at least 4 polys for 
	 a tube-shaped line segment*/
      number_of_polygons += 4; 
    }
    else if (strcmp(conntype,"triangles") == 0) {
      fprintf(f, "%s\t%d, %d, %d, -1,\n", tabstr, data->p, data->q, data->r);
      number_of_polygons += 1;
    }
    else if (strcmp(conntype,"quads") == 0) {
      fprintf(f, "%s\t%d, %d, %d, %d, -1,\n", 
	      tabstr, data->p, data->r, data->s, data->q);
      number_of_polygons += 1;
    }
    else if (strcmp(conntype,"tetrahedra") == 0) {
      fprintf(f, "%s\t%d, %d, %d, -1,\n", tabstr, data->p, data->q, data->r);
      fprintf(f, "%s\t%d, %d, %d, -1,\n", tabstr, data->q, data->s, data->r);
      fprintf(f, "%s\t%d, %d, %d, -1,\n", tabstr, data->s, data->p, data->r);
      fprintf(f, "%s\t%d, %d, %d, -1,\n", tabstr, data->p, data->q, data->s);
      number_of_polygons += 4;
    }
    else if (strcmp(conntype,"cubes") == 0) {
      fprintf(f, "%s\t%d, %d, %d, %d, -1,\n", 
	      tabstr, data->p, data->q, data->s, data->r);
      fprintf(f, "%s\t%d, %d, %d, %d, -1,\n", 
	      tabstr, data->q, data->u, data->w, data->s);
      fprintf(f, "%s\t%d, %d, %d, %d, -1,\n", 
	      tabstr, data->u, data->t, data->v, data->w);
      fprintf(f, "%s\t%d, %d, %d, %d, -1,\n", 
	      tabstr, data->t, data->p, data->r, data->v);
      fprintf(f, "%s\t%d, %d, %d, %d, -1,\n", 
	      tabstr, data->s, data->w, data->v, data->r);
      fprintf(f, "%s\t%d, %d, %d, %d, -1,\n", 
	      tabstr, data->q, data->u, data->t, data->p);
      number_of_polygons += 6;
    }
    else
      DXErrorReturn(ERROR_INTERNAL, "Connections are not of type lines, triangles, tetrahedra, or cubes");
  
  }
  
  fprintf(f, "%s]\n", tabstr);

  /* Now output the normal indices, one for each face (if normals are dependent
     on positions) or vertex (if normals are dependent on connections)
     */

  if (IgnoreNormals == FALSE) {
    
    /* if connections aren't lines, print the indices into normals */
    if (strcmp(conntype,"lines") != 0) { 
      /* shorthand for using same indices as coordIndex field */
      fprintf(f, "%snormalIndex\t-1\n", tabstr); 
    }  
  }
  

  /* Now output the material indices, one for each face (if colors are 
     dependent on positions) or vertex (if colors are dependent on 
     connections)
     */
  if (IgnoreMaterials == FALSE) {
    /* shorthand for using same indices as coordIndex field */
    fprintf(f, "%smaterialIndex\t-1\n", tabstr); 
  }

 skipconnectionstuff:
  
  tab(DEC);
  if (fprintf(f, "%s}\n",tabstr) < 0) return(BAD);
  
  return(GOOD);
}

/*---------------------------------------------------------------------*/

Error VRMLMaterial(FILE *f, Array colarray, Array opacarray, int *numcolors,
		   char *colorbinding, char *conntype, float ambient, 
		   float diffuse, float specular,int shininess, 
		   int IgnoreOpacities, int IgnoreRenderingProperties ) {

  
  ArrayHandle h, o; /* handles to the colors and opacities arrays */
  int nitems, rank, shape, i, j, range;
  Type type;
  Category category;
  Vector *data;
  Vector scratch;
  char bindstr[25];

 
  /* Output the color bindings */

  if (strcmp(conntype,"POINTSET") == 0) {
      DXMessage("POINTSET Found! Setting material binding to PER_VERTEX");
      strcpy(bindstr, "PER_VERTEX");
  } 
  else {
    if (strcmp(colorbinding,"connections") == 0) { 
      /* face colors */
      DXMessage("Setting material binding to PER_FACE");
      strcpy(bindstr, "PER_FACE");
    }
    else 
      if (strcmp(colorbinding,"positions") == 0) { 
	/* vertex colors */
	strcpy(bindstr, "PER_VERTEX_INDEXED");
	DXMessage("Setting material binding to PER_VERTEX_INDEXED");
      }
      else { 
	/* Colors bound to something else. Guess PER_FACE */
	strcpy(bindstr, "PER_FACE");
	DXMessage("Setting material binding to PER_FACE");
      }
  }
  
  if (fprintf(f, "%sMaterialBinding {\n", tabstr) < 0) return(BAD);
  tab(INC);
  fprintf(f, "%svalue\t%s\n", tabstr, bindstr);
  tab(DEC);
  if (fprintf(f, "%s}\n",tabstr) < 0) return(BAD);


  /* if connections are cubes/tetrahedra we need to output 6/4 colors
     for every color entry when colors are bound to connections */
  if (strcmp(colorbinding,"connections") == 0) {
    if (strcmp(conntype,"cubes") == 0) range = 6;
    else if (strcmp(conntype,"tetrahedra") == 0) range = 4;
    else range = 1;
  }
  else range = 1;
  
  if (fprintf(f, "%sMaterial {\n", tabstr) < 0) return(BAD);
  tab(INC);

  /* opacity (convert to transparency for VRML) */

  if ((opacarray != NULL) && (IgnoreOpacities == FALSE)) {
    
    if (DXGetArrayInfo(opacarray,&nitems,&type,&category,&rank,&shape) == NULL)
      DXErrorReturn(ERROR_UNEXPECTED,
		    "Could not get information about the array of opacities");
    
    if ((rank != 0) || (type != TYPE_FLOAT) || (category != CATEGORY_REAL)) 
      DXErrorReturn(ERROR_UNEXPECTED,
		    "Opacities are not an array of floating point values");
    
    /* check if all opacities are the same */
    if (DXQueryConstantArray(opacarray, &nitems, &scratch) != NULL) {
      DXMessage("Opacities are constant");
      nitems = range = 1;      
    }

    if ((o =  DXCreateArrayHandle(opacarray)) == NULL) return(ERROR);

    fprintf(f, "%stransparency [", tabstr);  
    for (i = 0; i < nitems; i++) {
      data = DXGetArrayEntry(o, i, &scratch);
      if (i) fprintf(f, "%s\t", tabstr);
      for (j = 0; j<range; j++) /* and it's a quad/tetra connection */
	fprintf(f, " %g,\n", 1 - (data->x));
    } 
    fprintf(f, "%s]\n", tabstr);
    
  } /* end -- if ((opacarray != NULL) && ...  */



  if (DXGetArrayInfo(colarray,&nitems,&type,&category,&rank,&shape) == NULL) 
    DXErrorReturn(ERROR_UNEXPECTED,
		  "Could not get information about the array of colors");
  
  DXMessage("Got %d colors", nitems);
  
  if (numcolors != NULL) *numcolors = nitems;
  
  if ((rank != 1) || (shape != 3) || 
      (type != TYPE_FLOAT) || (category != CATEGORY_REAL)) 
    DXErrorReturn(ERROR_UNEXPECTED,
		  "Colors are not an array of floating point 3-vectors");
  
  /* check if all colors are the same */
  if (DXQueryConstantArray(colarray, &nitems, &scratch) != NULL) {
								   DXMessage("Colors are constant");
								   nitems = range = 1;     
								 }
  
  if ((h =  DXCreateArrayHandle(colarray)) == NULL) return(ERROR);
  
  if (IgnoreRenderingProperties == FALSE) {
    
    /* ambient color */
    fprintf(f, "%sambientColor [", tabstr);  
    for (i = 0; i < nitems; i++) {
      data = DXGetArrayEntry(h, i, &scratch);
      if (i) fprintf(f, "%s\t", tabstr);
      for (j = 0; j<range; j++) 
	fprintf(f, " %g %g %g,\n", data->x * ambient,
		data->y * ambient, data->z * ambient);
    } 
    fprintf(f, "%s]\n", tabstr);
    
    /* diffuse color */
    fprintf(f, "%sdiffuseColor [", tabstr);  
    for (i = 0; i < nitems; i++) {
      data = DXGetArrayEntry(h, i, &scratch);
      if (i) fprintf(f, "%s\t", tabstr);
      for (j = 0; j<range; j++)
	fprintf(f, " %g %g %g,\n", data->x * diffuse,
		data->y * diffuse, data->z * diffuse);
    }
    fprintf(f, "%s]\n", tabstr);
    
    /* specular color */
    fprintf(f, "%sspecularColor [", tabstr);  
    for (i = 0; i < nitems; i++) {
      data = DXGetArrayEntry(h, i, &scratch);
      if (i) fprintf(f, "%s\t", tabstr);
      for (j = 0; j<range; j++)
	fprintf(f, " %g %g %g,\n", data->x * specular, 
		data->y * specular, data->z * specular);
    }  
    fprintf(f, "%s]\n", tabstr);
    
    /* shininess */
    fprintf(f, "%sshininess\t%f\n", tabstr, DXtoVRMLShininess(shininess));  
    
  } 
  else {			/* We ARE ignoring rendering properties */
	 DXMessage("We want colors, but NOT rendering properties. using diffuse");
	 fprintf(f, "%sdiffuseColor [", tabstr);  
	 for (i = 0; i < nitems; i++) {
	   data = DXGetArrayEntry(h, i, &scratch);
	   if (i) fprintf(f, "%s\t", tabstr);
	   for (j = 0; j<range; j++)
	     fprintf(f, " %g %g %g,\n", data->x, data->y, data->z);
	 } 
	 fprintf(f, "%s]\n", tabstr);
	 
       }
  

  tab(DEC);
  if (fprintf(f, "%s}\n",tabstr) < 0) return(BAD);
    
  DXMessage("End of Material node, connection type is %s", conntype);

  return(GOOD);
}

/*---------------------------------------------------------------------*/

Error VRMLShapeHints(FILE *f) {
  
  char vertexordering[] = "COUNTERCLOCKWISE";
  char shapetype[] = "UNKNOWN_SHAPE_TYPE";
  char facetype[] = "CONVEX";
  char creaseangle[] = CREASEANGLE;

  if (fprintf(f, "%sShapeHints {\n", tabstr) < 0) return(BAD);
  tab(INC);

  fprintf(f, "%svertexOrdering\t%s\n", tabstr, vertexordering);  
  fprintf(f, "%sshapeType\t%s\n", tabstr, shapetype);  
  fprintf(f, "%sfaceType\t%s\n", tabstr, facetype);  
  fprintf(f, "%screaseAngle\t%s\n", tabstr, creaseangle);  

  tab(DEC);
  if (fprintf(f, "%s}\n",tabstr) < 0) return(BAD);

  return(GOOD);
}

/*---------------------------------------------------------------------*/

Error VRMLDirectionalLight(FILE *f, Light light) {

  Vector direction;
  RGBColor color;
  char ison[] = "TRUE";
  float intensity = 1.0;

  if (DXGetObjectClass((Object)light) != CLASS_LIGHT) 
    DXErrorReturn(ERROR_BAD_CLASS, "Light input is not of class Light!");
  
  
  err = DXGetError();
  if (DXQueryDistantLight(light, &direction, &color) == NULL) {
    if (err != DXGetError()) { 
      return(ERROR); /* must have been an invalid light since error was set */
    }
    else { 
      /* since error wasn't set, DXQueryDistantLight() probably returned NULL 
	 because the light was an ambient light and not a distant point 
	 light */
      DXMessage("This is an ambient light; VRML can't handle it. Ignoring it");
      return(GOOD);
    }
  }

  /* it must have been a distant point light if we've gotten this far */
  DXMessage("Got a distant light");
  DXMessage("color %g %g %g  direction %g %g %g\n", 
	    color.r, color.g, color.b,direction.x, direction.y, direction.z);

  /* negate the direction of the light so it shines toward the origin*/
  direction.x = -direction.x;
  direction.y = -direction.y;
  direction.z = -direction.z;

  if (fprintf(f, "%sDirectionalLight {\n", tabstr) < 0) return(BAD);
  tab(INC);

  fprintf(f, "%son\t%s\n", tabstr, ison);  
  fprintf(f, "%sintensity\t%g\n", tabstr, intensity);  
  fprintf(f, "%scolor\t%g %g %g\n", tabstr, color.r, color.g, color.b);  
  fprintf(f, "%sdirection\t%g %g %g\n", 
	  tabstr, direction.x, direction.y, direction.z);  

  tab(DEC);
  if (fprintf(f, "%s}\n",tabstr) < 0) return(BAD);

  return(GOOD);

}


/*---------------------------------------------------------------------*/

Error WebspaceHints(FILE *f, Object *in) {
  
  if ((err = WSCameraHints(f, in)) != GOOD) return(err);
  if ((err = WSTitleHints(f, in)) != GOOD) return(err);

  return(GOOD);
  
}

/*---------------------------------------------------------------------*/

Error WSTitleHints(FILE *f, Object *in) {

  char *title;

  if (in[TAB_TITLE] != NULL) {
    if (DXExtractString(in[TAB_TITLE],&title) == NULL) 
      DXErrorReturn(ERROR_UNEXPECTED, "Couldn't extract title of scene")
  }
  else {
    DXMessage("No scene title was specified. Ignoring title.");
    return(GOOD);
  }

  if (fprintf(f, "%sDEF Title Info {\n", tabstr) < 0) return(BAD);
  tab(INC);

  fprintf(f, "%sstring\t\"%s\"\n", tabstr, title);  

  tab(DEC);
  if (fprintf(f, "%s}\n",tabstr) < 0) return(BAD);
 

  return(GOOD);
  
}

/*---------------------------------------------------------------------*/

Error WSCameraHints(FILE *f, Object *in) {
  
  char *viewname, *tempstr, s;
  char defaultname[16];
  int len, i, c;
  int tabnum;


  /* check to make sure at least ONE camera input is present */
  for (tabnum = TAB_CAMSTART; tabnum < MAX_NUM_TABS; tabnum += 2)
    if (in[tabnum] != NULL) goto gotvalidcam;
  DXMessage("No cameras are present");
  return(GOOD);

 gotvalidcam:



  /* remember how much we tabbed before */
  len = strlen(tabstr);

  /* reset tab to zero */
  tab(RESET);
  
  if (fprintf(f, "DEF Cameras Switch {\n") < 0) return(BAD);
  fprintf(f, "\twhichChild\t0\n");

  
  for (c = 1, tabnum = TAB_CAMSTART; tabnum < MAX_NUM_TABS; tabnum += 2) {
    if  (in[tabnum] != NULL) {
      /* Get the camera scene name */
      if (in[tabnum+1] != NULL) {
	if (DXExtractString(in[tabnum+1],&tempstr) == NULL) 
	  DXErrorReturn(ERROR_UNEXPECTED, "Couldn't extract name of viewpoint")
	else 
	  viewname = tempstr;
      }
      else {
	sprintf(defaultname, "ViewPoint(%d)", c++);
	viewname = defaultname;
	DXMessage("No viewpoint name was specified. Using default");
      }
      
      /* check the viewpoint name to make sure it's a valid node name */
      if (isdigit(viewname[0])) goto badnodename;
      i = 0;
      while (viewname[i] != '\0') {
	if (
	    iscntrl(viewname[i]) ||
	    isspace(viewname[i]) ||
	    (viewname[i] == '\'') ||
	    (viewname[i] == '"') ||
	    (viewname[i] == '\\') ||
	    (viewname[i] == '+') ||
	    (viewname[i] == '.') ||
	    (viewname[i] == '}') ||
	    (viewname[i] == '{')
	    ) 
	badnodename:
	  DXErrorReturn(ERROR_INVALID_DATA, "Viewpoint name must not begin with a number, and must not contain spaces, control characters, single or double quote characters, backslashes, curly braces, the plus character or the period character.");
	i++;
      }
          
      DXMessage("Viewpoint name will be \"%s\"", viewname);
      
      fprintf(f, "\tDEF %s ", viewname);
      if ((err = VRMLCamera(file, (Camera)in[tabnum])) != GOOD) return(err);
      
    }
  }
  
  if (fprintf(f, "%s}\n",tabstr) < 0) return(BAD);
  
  /* restore tab to what it was beforehand */
  tab(len);
    
  return(GOOD);
  
}

/*---------------------------------------------------------------------*/
